
/*
 * $Id: Client.cc,v 1.43.2.5 1999/10/05 19:58:49 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#include "polygraph.h"

#include "CltConnMgr.h"
#include "CltOpts.h"
#include "CltXact.h"
#include "EphPortMgr.h"
#include "ExpPortMgr.h"
#include "PopModel.h"
#include "ServerRep.h"
#include "StatPhase.h"
#include "StatPhaseMgr.h"
#include "LogComment.h"
#include "ErrorMgr.h"
#include "Client.h"

#include "RobotSym.h"

#include "polyBcastChannels.h"
#include "polyErrors.h"
#include "globals.h"


ObjFarm<CltXact> Client::TheXacts;
Array<PortMgr*> Client::ThePortMgrs;
Array<ServerRep*> Client::TheServers;
int Client:: TheShrdWorkSetCap = -1;


Client::Client(): theConnMgr(0), theOriginSel(0),
	theWaitXactLmt(-1), theCcXactLvl(0),
	theLaunchCount(0), theExtraLaunchLvl(0),
	thePopModel(0),
	thePublicRatio(-1), theRecurRatio(-1), theEmbedRecurRatio(-1),
	genUniqUrls(false) {

	theChannels.append(TheInfoChannel);
	startListen();
}

Client::~Client() {
}

void Client::configure(const RobotSym *cfg, const NetAddr &aHost, bool bind2iface) {
	SockOpt opt;
	Agent::configure(cfg, aHost, opt);
	theHost.port(-1); // we do not care about port number

	int openConnLmt = -1;
	cfg->openConnLimit(openConnLmt);
	if (!cfg->waitXactLimit(theWaitXactLmt) && openConnLmt >= 0)
		theWaitXactLmt = 2*openConnLmt; // default

	const int srvCnt = scanServers(cfg);
	theConnMgr = new CltConnMgr;
	theConnMgr->configure(opt, cfg->pconnUseLmt(), openConnLmt,
		TheCltOpts.thePxyAddr ? 1 : srvCnt);
	theConnMgr->portMgr(getPortMgr(bind2iface));

	theLaunchWin = cfg->launchWin();
	cfg->publicRatio(thePublicRatio);
	cfg->recurRatio(theRecurRatio);
	cfg->embedRecurRatio(theEmbedRecurRatio);
	thePopModel = cfg->popModel();
	cfg->uniqueUrls(genUniqUrls);

	if (!thePopModel && theRecurRatio > 0) {
		cerr << cfg->loc() << "popularity model must be specified for"
			<< " positive recurrence ratio (robot " << theKind << ')'
			<< endl;
		exit(-2);
	}	

	TheXacts.limit(1024); // magic, no good way to estimate
}

void Client::start() {
	Assert(!theLaunchCount);
	Assert(theConnMgr);

	Agent::start();

	if (theLaunchWin > 0) {
		// uniformly distributed start delay
		static RndGen rng;
		const double delay = rng(0.0, theLaunchWin.secd());
		sleepFor(Time::Secd(delay));
	} else {
		scheduleLaunch(TheClock);
	}
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWorkSetFreeze) {
		for (int i = 0; i < thePrivWorkSetCap.count(); ++i)
			thePrivWorkSetCap[i] = thePrivOids[i];

		if (TheShrdWorkSetCap < 0)
			FreezeSrvWorkSetCap();
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);
	ConnMgr *mgr = x->conn()->mgr();
	Assert(mgr);
	mgr->put(x->conn());
	TheXacts.put(x);
	theCcXactLvl--;
	Assert(theCcXactLvl >= 0);

	// push waiting xactions forward
	if (theLaunchDebts.count() && !theConnMgr->atHardConnLimit())
		launch(theLaunchDebts.pop());
}

void Client::scheduleLaunch(Time lastLaunch) {
	Assert(!thePendAlarmCnt);
	if (theThinkDistr) {
		const double lf = TheStatPhaseMgr->loadFactor(lastLaunch);
		Assert(lf > 0);
		double delay;
		if (TheStatPhaseMgr->varyLoad()) {
			// next launch should happen [0, 2dt] from last launch
			// where dt is 1/current_rate
			static RndGen rng;
			delay = rng(0.0, 2.*theThinkDistr->mean()/lf);
		} else {
			delay = theThinkDistr->trial()/lf;
		}
		sleepTill(lastLaunch + Time::Secd(delay));
	} else {
		launch();
	}
}

bool Client::launch(CltXact *x) {
	// connect and execute the transaction
	const NetAddr &connAddr = TheCltOpts.thePxyAddr ?
		(const NetAddr &)TheCltOpts.thePxyAddr : 
		(const NetAddr &)(GetServerRep(x->oidRec().sidx)->addr());

	if (Connection *conn = theConnMgr->get(connAddr)) {
		theLaunchCount++;
		theCcXactLvl++;
		x->exec(this, conn);
		return true;
	}

	if (ReportError(errno))
		Comment(1) << theHost << " failed to connect to " << connAddr << endc;
	Broadcast(TheXactErrChannel, x);

	TheXacts.put(x);

	// if no alarms are pending, we may get stuck and submit
	// no more xactions (e.g. best effort); this is a kludge
	if (!thePendAlarmCnt)
		sleepTill(TheClock); // submit more next cycle

	return false;
}

bool Client::launch(const OidRec &or) {
	CltXact *x = TheXacts.get();
	x->oidRec(or);

	// check if we should postpone the xaction
	if (theConnMgr->atHardConnLimit()) {
		if (theWaitXactLmt < 0 || theLaunchDebts.count() < theWaitXactLmt) {
			theLaunchDebts.append(x);
			return true;
		}

		if (ReportError(errTooManyWaitXact))
			Comment(3) << "xactions active: " << theCcXactLvl 
				<< " waiting: " << theLaunchDebts.count() 
				<< " limit: " << theWaitXactLmt << endc;
		TheXacts.put(x);
		return false;
	}

	return launch(x);
}

void Client::launch() {
	OidRec or;
	or.sidx = (int)theOriginSel->trial();
	genOid(or);
	or.tid = 0; // direct access
	launch(or);
}

void Client::genOid2(OidRec &or) {
	ServerRep &srvRep = *GetServerRep(or.sidx);

	static RndGen rng;
	const bool publicOid = rng.event(thePublicRatio);
	const bool repeatOid = rng.event(theRecurRatio);

	if (publicOid && srvRep.id()) {
		or.wid = srvRep.id(); // use server id

		// can repeat?
		if (repeatOid && srvRep.secOid() > 0) {
			or.repeat = true;
			or.oid = thePopModel->select(rng, srvRep.secOid(), srvRep.workSetCap());
			return;
		}

		// can ask for a new oid?
		if (srvRep.haveNewOids()) {
			or.oid = srvRep.newOid();
			return;
		}

		// report insufficient number of new oids if server is initialized
		if (srvRep.secOid() > srvRep.maxNewOidCap())
			ReportError(errNoNewOids);

		// fall through if cannot generate a public oid
	}
		
	// ask for a private object
	or.wid = theId; // use this robot id

	// can repeat?
	int &privOid = thePrivOids[or.sidx];
	if (repeatOid && privOid > 0) {
		or.repeat = true;
		or.oid = thePopModel->select(rng, privOid, thePrivWorkSetCap[or.sidx]);
		return;
	}

	// last resort: new private object
	or.oid = ++privOid;
}

void Client::genOid(OidRec &or) {
	genOid2(or);
	if (genUniqUrls)
		or.wid = UniqId::Create(); // changes every time
}

void Client::noteEmbedded(CltXact *, const OidRec &or) {
	static RndGen rng;
	if (or.repeat && !rng.event(theEmbedRecurRatio))
		return; // ignore this embedded object as if we had it cached

	if (launch(or))
		theExtraLaunchLvl++;
}

int Client::scanServers(const RobotSym *cfg) {
	Array<NetAddr*> addrs;
	if (!cfg->origins(addrs, theOriginSel) || !addrs.count()) {
		Comment(1) << "no origin addresses specified for robot " << cfg->kind() << endc;
		FatalError(errOther);
	}
	Assert(theOriginSel);

	thePrivOids.stretch(addrs.count());
	thePrivWorkSetCap.stretch(addrs.count());
	for (int i = 0; i < addrs.count(); ++i) {
		thePrivOids.append(0);
		thePrivWorkSetCap.append(-1);
	}

	const int savedCount = TheServers.count();
	for (int i = 0; i < addrs.count(); ++i)
		cAddServer(*addrs[i]);

	if (savedCount && TheServers.count() != savedCount) {
		Comment(1) << "varying the number of origin servers per robot "
			<< "is not yet supported for memory consumption reasons; "
			<< "let us know if you need this feature" << endl;
		FatalError(errOther);
	}

	// get rid of addresses
	while (addrs.count()) delete addrs.pop();
	return TheServers.count();
}

PortMgr *Client::getPortMgr(bool bind2iface) {
	const NetAddr addr = bind2iface ? theHost : NetAddr();

	// check if we already have a port mgr for host
	for (int i = 0; i < ThePortMgrs.count(); ++i) {
		if (ThePortMgrs[i]->addr() == addr)
			return ThePortMgrs[i];
	}

	// create new port manager
	PortMgr *mgr = TheCltOpts.thePorts.set() ?
		(PortMgr*) new ExpPortMgr(addr, TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		(PortMgr*) new EphPortMgr(addr);
	ThePortMgrs.append(mgr);
	return mgr;
}

// creates new server representative if needed
// returns true iff creation was needed
bool Client::cAddServer(const NetAddr &addr) {
	// check if we already know about this server
	for (int i = 0; i < TheServers.count(); ++i) {
		if (TheServers[i]->addr() == addr)
			return false;
	}
	
	// new server
	TheServers.append(new ServerRep(addr));
	return true;
}

ServerRep *Client::GetServerRep(int idx) {
	Assert(0 <= idx && idx < TheServers.count());
	return TheServers[idx];
}

void Client::FreezeSrvWorkSetCap() {
	Assert(TheShrdWorkSetCap < 0);

	TheShrdWorkSetCap = 0;
	for (int i = 0; i < TheServers.count(); ++i) {
		TheServers[i]->freezeWorkSetCap();
		TheShrdWorkSetCap += TheServers[i]->workSetCap();
	}
}
